@@ -3,7 +3,7 @@
 import sys
 sys.path.append("debian/lib/python")
 
-import codecs
+import locale
 import errno
 import glob
 import io
@@ -11,6 +11,8 @@ import os
 import os.path
 import subprocess
 
+locale.setlocale(locale.LC_CTYPE, "C.UTF-8")
+
 from debian_linux import config
 from debian_linux.debian import *
 from debian_linux.gencontrol import Gencontrol as Base, merge_packages
@@ -23,7 +25,7 @@ class Gencontrol(Base):
         },
         'build': {
             'debug-info': config.SchemaItemBoolean(),
-            'signed-modules': config.SchemaItemBoolean(),
+            'signed-code': config.SchemaItemBoolean(),
             'vdso': config.SchemaItemBoolean(),
         },
         'description': {
@@ -35,6 +37,7 @@ class Gencontrol(Base):
             'initramfs-generators': config.SchemaItemList(),
             'check-size': config.SchemaItemInteger(),
             'check-size-with-dtb': config.SchemaItemBoolean(),
+            'check-uncompressed-size': config.SchemaItemInteger(),
         },
         'relations': {
         },
@@ -43,7 +46,9 @@ class Gencontrol(Base):
             'headers-all': config.SchemaItemBoolean(),
             'installer': config.SchemaItemBoolean(),
             'libc-dev': config.SchemaItemBoolean(),
-            'tools': config.SchemaItemBoolean(),
+            'tools-unversioned': config.SchemaItemBoolean(),
+            'tools-versioned': config.SchemaItemBoolean(),
+            'source': config.SchemaItemBoolean(),
         }
     }
 
@@ -61,7 +66,7 @@ class Gencontrol(Base):
                 makeflags[dst] = data[src]
 
     def _substitute_file(self, template, vars, target, append=False):
-        with codecs.open(target, 'a' if append else 'w', 'utf-8') as f:
+        with open(target, 'a' if append else 'w') as f:
             f.write(self.substitute(self.templates[template], vars))
 
     def do_main_setup(self, vars, makeflags, extra):
@@ -72,6 +77,10 @@ class Gencontrol(Base):
             'ABINAME': self.abiname_version + self.abiname_part,
             'SOURCEVERSION': self.version.complete,
         })
+        if not self.config.merge('packages').get('tools-unversioned', True):
+            makeflags['DO_TOOLS_UNVERSIONED'] = False
+        if not self.config.merge('packages').get('tools-versioned', True):
+            makeflags['DO_TOOLS_VERSIONED'] = False
 
         # Prepare to generate debian/tests/control
         self.tests_control = None
@@ -101,10 +110,11 @@ class Gencontrol(Base):
         makeflags = makeflags.copy()
         makeflags['ALL_FEATURESETS'] = ' '.join(fs_enabled)
         makeflags['ALL_TRIPLETS'] = ' '.join(triplet_enabled)
+        makeflags['SOURCE_BASENAME'] = self.vars['source_basename']
         if not self.config.merge('packages').get('docs', True):
             makeflags['DO_DOCS'] = False
-        if not self.config.merge('packages').get('tools', True):
-            makeflags['DO_TOOLS'] = False
+        if not self.config.merge('packages').get('source', True):
+            makeflags['DO_SOURCE'] = False
         super(Gencontrol, self).do_main_makefile(makefile, makeflags, extra)
 
         # linux-source-$UPSTREAMVERSION will contain all kconfig files
@@ -114,10 +124,12 @@ class Gencontrol(Base):
         packages.extend(self.process_packages(self.templates["control.main"], self.vars))
         if self.config.merge('packages').get('docs', True):
             packages.extend(self.process_packages(self.templates["control.docs"], self.vars))
-        if self.config.merge('packages').get('tools', True):
-            packages.extend(self.process_packages(self.templates["control.tools"], self.vars))
+        if self.config.merge('packages').get('tools-unversioned', True):
+            packages.extend(self.process_packages(self.templates["control.tools-unversioned"], self.vars))
+        if self.config.merge('packages').get('tools-versioned', True):
+            packages.extend(self.process_packages(self.templates["control.tools-versioned"], self.vars))
 
-        self._substitute_file('lintian-overrides.perf', self.vars,
+        self._substitute_file('perf.lintian-overrides', self.vars,
                               'debian/linux-perf-%s.lintian-overrides' %
                               self.vars['version'])
 
@@ -174,6 +186,8 @@ class Gencontrol(Base):
             makeflags['ABINAME'] = vars['abiname'] = \
                 self.abiname_version + abiname_part
 
+        build_signed = self.config.merge('build', arch).get('signed-code', False)
+
         # Some userland architectures require kernels from another
         # (Debian) architecture, e.g. x32/amd64.
         # And some derivatives don't need the headers-all packages
@@ -194,10 +208,6 @@ class Gencontrol(Base):
         else:
             makeflags['DO_LIBC'] = False
 
-        if not self.config.merge('packages').get('tools', True):
-            makeflags['DO_TOOLS'] = False
-
-
         merge_packages(packages, packages_headers_arch, arch)
 
         cmds_build_arch = ["$(MAKE) -f debian/rules.real build-arch-arch %s" % makeflags]
@@ -223,11 +233,6 @@ class Gencontrol(Base):
             installer_def_dir = 'debian/installer'
             installer_arch_dir = os.path.join(installer_def_dir, arch)
             if os.path.isdir(installer_arch_dir):
-                # If we're going to build signed udebs later, don't actually
-                # generate udebs.  Just test that we *can* build, so we find
-                # configuration errors before building linux-signed.
-                test_build = self.config.merge('build', arch).get('signed-modules', False)
-
                 kw_env = os.environ.copy()
                 kw_env['KW_DEFCONFIG_DIR'] = installer_def_dir
                 kw_env['KW_CONFIG_DIR'] = installer_arch_dir
@@ -236,20 +241,30 @@ class Gencontrol(Base):
                     stdout=subprocess.PIPE,
                     env=kw_env)
                 if not isinstance(kw_proc.stdout, io.IOBase):
-                    udeb_packages = read_control(io.open(kw_proc.stdout.fileno(), encoding='utf-8', closefd=False))
+                    udeb_packages = read_control(io.open(kw_proc.stdout.fileno(), closefd=False))
                 else:
-                    udeb_packages = read_control(io.TextIOWrapper(kw_proc.stdout, 'utf-8'))
+                    udeb_packages = read_control(io.TextIOWrapper(kw_proc.stdout))
                 kw_proc.wait()
                 if kw_proc.returncode != 0:
                     raise RuntimeError('kernel-wedge exited with code %d' %
                                        kw_proc.returncode)
 
+                # If we're going to build signed udebs later, don't actually
+                # generate udebs.  Just test that we *can* build, so we find
+                # configuration errors before building linux-signed.
+
                 # kernel-wedge currently chokes on Build-Profiles so add it now
                 for package in udeb_packages:
-                    package['Build-Profiles'] = '<!stage1>'
+                    if build_signed:
+                        # XXX This is a hack to exclude the udebs from
+                        # the package list while still being able to
+                        # convince debhelper and kernel-wedge to go
+                        # part way to building them.
+                        package['Build-Profiles'] = '<pkg.linux.udeb-unsigned-test-build>'
+                    else:
+                        package['Build-Profiles'] = '<!stage1>'
 
-                if not test_build:
-                    merge_packages(packages, udeb_packages, arch)
+                merge_packages(packages, udeb_packages, arch)
 
                 # These packages must be built after the per-flavour/
                 # per-featureset packages.  Also, this won't work
@@ -261,7 +276,19 @@ class Gencontrol(Base):
                               "PACKAGE_NAMES='%s' UDEB_UNSIGNED_TEST_BUILD=%s" %
                               (arch, makeflags,
                                ' '.join(p['Package'] for p in udeb_packages),
-                               test_build)])
+                               build_signed)])
+
+        # This also needs to be built after the per-flavour/per-featureset
+        # packages.
+        if build_signed:
+            merge_packages(packages,
+                           self.process_packages(
+                               self.templates['control.signed-template'], vars),
+                           arch)
+            makefile.add(
+                'binary-arch_%s' % arch,
+                cmds=["$(MAKE) -f debian/rules.real install-signed-template_%s %s" %
+                      (arch, makeflags)])
 
     def do_featureset_setup(self, vars, makeflags, arch, featureset, extra):
         config_base = self.config.merge('base', arch, featureset)
@@ -269,6 +296,7 @@ class Gencontrol(Base):
 
     flavour_makeflags_base = (
         ('compiler', 'COMPILER', False),
+        ('compiler-filename', 'COMPILER', True),
         ('kernel-arch', 'KERNEL_ARCH', False),
         ('cflags', 'CFLAGS_KERNEL', True),
         ('override-host-type', 'OVERRIDE_HOST_TYPE', True),
@@ -337,7 +365,7 @@ class Gencontrol(Base):
         for group in relations_compiler_build_dep:
             for item in group:
                 item.arches = [arch]
-        packages['source']['Build-Depends'].extend(relations_compiler_build_dep)
+        packages['source']['Build-Depends-Arch'].extend(relations_compiler_build_dep)
 
         image_fields = {'Description': PackageDescription()}
         for field in 'Depends', 'Provides', 'Suggests', 'Recommends', 'Conflicts', 'Breaks':
@@ -383,7 +411,7 @@ class Gencontrol(Base):
         packages_dummy = []
         packages_own = []
 
-        build_signed = config_entry_build.get('signed-modules')
+        build_signed = config_entry_build.get('signed-code')
 
         image = self.templates[build_signed and "control.image-unsigned"
                                or "control.image"]
@@ -498,9 +526,11 @@ class Gencontrol(Base):
             self._substitute_file('image.%s' % name, vars,
                                   'debian/%s.%s' % (image_main['Package'], name))
         if build_debug:
-            self._substitute_file('image-dbg.lintian-override', vars,
-                                  'debian/linux-image-%s%s-dbg.lintian-overrides' %
+            debug_lintian_over = ('debian/linux-image-%s%s-dbg.lintian-overrides' %
                                   (vars['abiname'], vars['localversion']))
+            self._substitute_file('image-dbg.lintian-overrides', vars,
+                                  debug_lintian_over)
+            os.chmod(debug_lintian_over, 0o755)
 
     def process_changelog(self):
         act_upstream = self.changelog[0].version.upstream
@@ -522,6 +552,7 @@ class Gencontrol(Base):
         self.vars = {
             'upstreamversion': self.version.linux_upstream,
             'version': self.version.linux_version,
+            'source_basename': re.sub(r'-[\d.]+$', '', self.changelog[0].source),
             'source_upstream': self.version.upstream,
             'source_package': self.changelog[0].source,
             'abiname': self.abiname_version + self.abiname_part,
@@ -569,7 +600,7 @@ class Gencontrol(Base):
         f.close()
 
     def write_tests_control(self):
-        self.write_rfc822(codecs.open("debian/tests/control", 'w', 'utf-8'),
+        self.write_rfc822(open("debian/tests/control", 'w'),
                           [self.tests_control])
 
 if __name__ == '__main__':
